
#include "gsconvert.hpp"
#include "winutil.hpp"
#include <cstddef>
#include <cstdlib>
#include <algorithm>
#include <stdexcept>

#include "fract.hpp"


namespace agraph
{

colortograyscalefactor_t& colortograyscalefactor_t::setToDefault (void)
{
       blue_factor = 0.11f;
       green_factor = 0.59f;
       red_factor = 0.30f;

       return *this;
}
void colortograyscalefactor_t::normalize (void)
{
    float sum = blue_factor + green_factor + red_factor;
    if ((sum < 0.9999) | (sum > 1.0001))
    {
        blue_factor /= sum;
        green_factor /= sum;
        red_factor /= sum;
    }
}

unsigned colortable_t::findNearestColor (float rgb) const
{
    const DWORD* p = reinterpret_cast <const DWORD*> (rgbvals);
    unsigned c = colors - 1;

    if (rgb > (static_cast <float> (p[c] & 0xFF) + 1))
        return c;

    for (int n = colors - 2; n >= 0; --n)
        if (rgb > (static_cast <float> (p[n] & 0xFF) + 1))
            return ((rgb - p[n]) < (rgb - p[n + 1])) ? n : n + 1;

    return 0;
}
unsigned colortable_t::findNearestColor (unsigned char rgb) const
{
    const DWORD* p = reinterpret_cast <const DWORD*> (rgbvals);
    unsigned c = colors - 1;

    if (rgb >= static_cast <unsigned char> (p[c] & 0xFF))
        return c;

    for (int n = colors - 2; n >= 0; --n)
        if (rgb >= static_cast <unsigned char> (p[c] & 0xFF))
            return ((rgb - p[n]) < (rgb - p[n + 1])) ? n : n + 1;

    return 0;
}

template <int Min, int Max, typename T>
    T bound (T val)
{ return (val >= Min) ? ((val <= Max) ? val : Max) : Min; }


ucl::stl::ArrayPtr <colortable_t> colortable_t::createColorTable (unsigned char colors)
{
    char* buf;

    buf = new char[offsetof (colortable_t, rgbvals) + sizeof (RGBQUAD) * colors];
    new (reinterpret_cast <colortable_t*> (buf)) colortable_t (colors);
    return ucl::stl::ArrayPtr <colortable_t> (reinterpret_cast <colortable_t*> (buf));
}

colortable_t& colortable_t::setToDefault (void)
{
    float diff = 255.0 / (colors - 1);
    float clr = 0.5;
    RGBQUAD* p;
    RGBQUAD* c;

    for (p = rgbvals, c = p + colors; p < c; ++p, clr += diff)
    {
        p->rgbBlue = p->rgbGreen = p->rgbRed = static_cast <unsigned char> (clr);
        p->rgbReserved = 0;
    }

    return *this;
}



GSBitmap::GSBitmap (void)
{}
GSBitmap::GSBitmap (const GSBitmap& rhs)
{
    obmdata.copy (rhs.obmdata);
    obminfo.copy (rhs.obminfo);
}
GSBitmap& GSBitmap::operator = (const GSBitmap& rhs)
{
    obmdata.copy (rhs.obmdata);
    obminfo.copy (rhs.obminfo);
    return *this;
}
GSBitmap::~GSBitmap (void)
{}


void GSBitmap::pasteFromBitmap (const BITMAPINFOHEADER* bmih, const BYTE* bdata,
        checkfile_t cf, void* cfdata)
{
    static const unsigned char pshifttbl[] = {0, 3, 2, 0, 1, 0, 0, 0, 0};

    static const unsigned pshift2tbl1[] = {7, 6, 5, 4, 3, 2, 1, 0};
    static const unsigned pshift2tbl2[] = {6, 4, 2, 0, 6, 4, 2, 0};
    static const unsigned pshift2tbl4[] = {4, 0, 4, 0, 4, 0, 4, 0};
    static const unsigned pshift2tbl8[] = {0, 0, 0, 0, 0, 0, 0, 0};
    static const unsigned* pshift2tbls[] = {0,
                                            pshift2tbl1, pshift2tbl2,
                                            0,           pshift2tbl4,
                                            0,           0,
                                            0,           pshift2tbl8};


    if (!(bmih && bdata))
        return;

    const RGBQUAD* ctbl = reinterpret_cast <const RGBQUAD*>
        (reinterpret_cast <const char*> (bmih) + bmih->biSize);

    unsigned srcrowlen, dstrowlen;


    if (bmih->biCompression)
        throw std::runtime_error ("GSBitmap: RLE compression not yet supported.");

    if (cf)
        cf (cfdata, bmih);

    srcrowlen = ((bmih->biWidth * bmih->biBitCount + 31) & ~31) >> 3;
    dstrowlen = ((bmih->biWidth * 24 + 31) & ~31) >> 3;

    obminfo.alloc (1); // no color table needed
    obminfo->bmiHeader.biSize = sizeof (obminfo->bmiHeader);
    obminfo->bmiHeader.biWidth = bmih->biWidth;
    obminfo->bmiHeader.biHeight = bmih->biHeight;
    obminfo->bmiHeader.biPlanes = 1;
    obminfo->bmiHeader.biBitCount = 24;
    obminfo->bmiHeader.biCompression = BI_RGB;
    obminfo->bmiHeader.biSizeImage = dstrowlen * bmih->biHeight;
    obminfo->bmiHeader.biXPelsPerMeter = obminfo->bmiHeader.biYPelsPerMeter = 0;
    obminfo->bmiHeader.biClrUsed = 0;
    obminfo->bmiHeader.biClrImportant = 0;

    obmdata.alloc (obminfo->bmiHeader.biSizeImage + 4);

    if (bmih->biBitCount == 24) // no need to transform anything, just copy
        std::memcpy (obmdata.get (), bdata, obminfo->bmiHeader.biSizeImage);
    else
    { // transform to 24 bpp bitmap
        const unsigned char* src;
        unsigned char* dst;
        const unsigned char* xsrc;
        unsigned char* xdst;
        DWORD rgb;
        bool usetbl = (bmih->biBitCount <= 8);
        unsigned psz = bmih->biBitCount >> 3;
        unsigned pmask = (1 << bmih->biBitCount) - 1;
        unsigned pshift = usetbl ? pshifttbl[bmih->biBitCount] : 0;
        const unsigned* poffstbl = usetbl ? pshift2tbls[bmih->biBitCount] : 0;

        src = bdata;
        dst = obmdata.get ();

        if (usetbl)
        {
            for (int y = 0; y < bmih->biHeight; ++y)
            {
                xdst = dst;
                for (int x = 0; x < bmih->biWidth; ++x)
                {
                    *reinterpret_cast <RGBQUAD*> (xdst)
                        = ctbl[(src[x >> pshift] >> poffstbl[x & 7]) & pmask];
                    xdst += 3;
                }
                dst += dstrowlen;
                src += srcrowlen;
            }
        }
        else if (bmih->biBitCount == 32)
        {
            for (int y = 0; y < bmih->biHeight; ++y)
            {
                xdst = dst;
                xsrc = src;
                for (int x = 0; x < bmih->biWidth; ++x)
                {
                    *reinterpret_cast <RGBQUAD*> (xdst)
                        = *(reinterpret_cast <const RGBQUAD*> (xsrc)) /* & pmask */;
                    xdst += 3;
                    xsrc += psz;
                }
                dst += dstrowlen;
                src += srcrowlen;
            }
        }
        else
            throw std::runtime_error ("GSBitmap: Bit width not supported!");
    }
}

void GSBitmap::loadFromBitmap (const char* filename, checkfile_t cf, void* cfdata)
{
    ucl::win::WindowsHandle <HANDLE> hFile;
    BITMAPFILEHEADER bmfh;
    ucl::stl::ArrayPtr <char> binf;
    ucl::stl::ArrayPtr <char> pbuf;
    const BITMAPINFOHEADER* bmih;
    DWORD nobr;


    hFile = CreateFile (filename, GENERIC_READ, FILE_SHARE_READ,
        NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (!hFile.isValid ())
        throw std::runtime_error ("GSBitmap: Unable to open file!");

    ReadFile (hFile, &bmfh, sizeof (bmfh), &nobr, NULL);

    if ((bmfh.bfType != 0x4D42) // "BM"
     || (nobr != sizeof (bmfh)))
        throw std::runtime_error ("GSBitmap: File is not a valid bitmap file!");

    binf.alloc (bmfh.bfOffBits - sizeof (bmfh));
    ReadFile (hFile, binf.get (), binf.size (), &nobr, NULL);

    bmih = reinterpret_cast <const BITMAPINFOHEADER*> (binf.get ());

    pbuf.alloc (bmih->biSizeImage);
    ReadFile (hFile, pbuf.get (), pbuf.size (), &nobr, NULL);

    pasteFromBitmap (bmih, pbuf.get (), cf, cfdata);
}

void GSBitmap::saveToBitmap (const char* filename)
{
    ucl::win::WindowsHandle <HANDLE> hFile;
    BITMAPFILEHEADER          bmfh;
    BITMAPINFOHEADER          bmih;
    ucl::stl::ArrayPtr <unsigned char> pbuf;
    const unsigned char*      src;
    unsigned char*            dst;
    unsigned                  srcrowlen;
    unsigned                  dstrowlen;
    DWORD                     nobw;
    unsigned                  bpp;


    if (!obmdata.get ())
        return;

    hFile = CreateFile (filename, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS,
        FILE_ATTRIBUTE_NORMAL, NULL);
    if (!hFile.isValid ())
        throw std::runtime_error ("GSBitmap: Unable to write file!");

    if (obminfo->bmiHeader.biBitCount == 8)
    {
        if (obminfo->bmiHeader.biClrUsed <= 2)
            bpp = 1;
        else if (obminfo->bmiHeader.biClrUsed <= 16)
            bpp = 4;
        else if (obminfo->bmiHeader.biClrUsed <= 256)
            bpp = 8;
    }
    else
        bpp = 24;


    srcrowlen = ((obminfo->bmiHeader.biWidth * 8 + 31) & ~31) >> 3;
    dstrowlen = ((obminfo->bmiHeader.biWidth * bpp + 31) & ~31) >> 3;

    bmfh.bfType = 0x4D42; // "BM"
    bmfh.bfSize = sizeof (BITMAPFILEHEADER)
                + sizeof (BITMAPINFOHEADER)
                + sizeof (RGBQUAD) * obminfo->bmiHeader.biClrUsed
                + dstrowlen * obminfo->bmiHeader.biHeight;
    bmfh.bfReserved1 = bmfh.bfReserved2 = 0;
    bmfh.bfOffBits = sizeof (BITMAPFILEHEADER)
                + sizeof (BITMAPINFOHEADER)
                + sizeof (RGBQUAD) * obminfo->bmiHeader.biClrUsed;

    WriteFile (hFile, &bmfh, sizeof (bmfh), &nobw, NULL);

    if (bpp >= 8)
    {
        WriteFile (hFile, obminfo.get (),
            sizeof (BITMAPINFOHEADER) + sizeof (RGBQUAD) * obminfo->bmiHeader.biClrUsed,
            &nobw, NULL);
        WriteFile (hFile, obmdata.get (), obmdata.size (), &nobw, NULL);
    }
    else
    {
        bmih.biSize = sizeof (bmih);
        bmih.biWidth = obminfo->bmiHeader.biWidth;
        bmih.biHeight = obminfo->bmiHeader.biHeight;
        bmih.biPlanes = 1;
        bmih.biBitCount = bpp;
        bmih.biCompression = 0;
        bmih.biSizeImage = dstrowlen * bmih.biHeight;
        bmih.biXPelsPerMeter = bmih.biYPelsPerMeter = 0;
        bmih.biClrUsed = obminfo->bmiHeader.biClrUsed;
        bmih.biClrImportant = obminfo->bmiHeader.biClrUsed;

        pbuf.alloc (bmih.biSizeImage + 1);
        std::memset (pbuf.get (), 0, pbuf.size ());

        dst = pbuf.get ();
        src = obmdata.get ();

        if (bpp == 1)
        {
            for (int y = 0; y < bmih.biHeight; ++y)
            {
                for (int x = 0; x < bmih.biWidth; ++x)
                    dst[x >> 3] |= src[x] << (7 - (x & 7));
                src += srcrowlen;
                dst += dstrowlen;
            }
        }
        else
        { // bpp == 4
            for (int y = 0; y < bmih.biHeight; ++y)
            {
                for (int x = 0; x < bmih.biWidth; ++x)
                    dst[x >> 1] |= src[x] << ((~x & 1) << 2);
                src += srcrowlen;
                dst += dstrowlen;
            }
        }

        WriteFile (hFile, &bmih, sizeof (bmih), &nobw, NULL);
        WriteFile (hFile, obminfo->bmiColors, sizeof (RGBQUAD) * bmih.biClrUsed, &nobw, NULL);
        WriteFile (hFile, pbuf.get (), dstrowlen * bmih.biHeight, &nobw, NULL);
    }
}


static DWORD averageRGB (DWORD rgb00, DWORD rgb10, DWORD rgb01, DWORD rgb11,
    const double& f00, const double& f10, const double& f01, const double& f11)
{
    unsigned char b = (rgb00 & 0xFF) * f00
                    + (rgb10 & 0xFF) * f10
                    + (rgb01 & 0xFF) * f01
                    + (rgb11 & 0xFF) * f11;
    unsigned char g = ((rgb00 >> 8) & 0xFF) * f00
                    + ((rgb10 >> 8) & 0xFF) * f10
                    + ((rgb01 >> 8) & 0xFF) * f01
                    + ((rgb11 >> 8) & 0xFF) * f11;
    unsigned char r = ((rgb00 >> 16) & 0xFF) * f00
                    + ((rgb10 >> 16) & 0xFF) * f10
                    + ((rgb01 >> 16) & 0xFF) * f01
                    + ((rgb11 >> 16) & 0xFF) * f11;
    return b | (g << 8) | (r << 16);
}

static inline BYTE averageGS (BYTE gs00, BYTE gs10, BYTE gs01, BYTE gs11,
    const double& f00, const double& f10, const double& f01, const double& f11)
{
    return (gs00 * f00
          + gs10 * f10
          + gs01 * f01
          + gs11 * f11);
}

void GSBitmap::performResize (unsigned width, unsigned height, resize_t resize)
{
    unsigned srcrowlen, dstrowlen;
    ucl::stl::ArrayPtr <BYTE> nbmdata;
    double xstep, ystep;
    double xpos, ypos;
    double xf1, xf2;
    double yf1, yf2;

    unsigned xl1, xl2;
    unsigned yl1, yl2;

    unsigned char* dst;
    unsigned char* xdst;
    const unsigned char* src;
    const unsigned char* ysrc1;
    const unsigned char* ysrc2;
    //const unsigned char* srcp;


    if (!getBitmap ())
        return;

    srcrowlen = ((obminfo->bmiHeader.biWidth * obminfo->bmiHeader.biBitCount + 31) & ~31) >> 3;
    dstrowlen = ((width * obminfo->bmiHeader.biBitCount + 31) & ~31) >> 3;

    xstep = static_cast <double> (obminfo->bmiHeader.biWidth) / width;
    ystep = static_cast <double> (obminfo->bmiHeader.biHeight) / height;

    nbmdata.alloc (dstrowlen * height + 4);
    dst = nbmdata.get ();
    src = obmdata.get ();

    if (obminfo->bmiHeader.biBitCount == 8)
    {
        switch (resize)
        {
        case reBilinear:
            ypos = 0.0;
            for (unsigned y = 0; y < height; ++y)
            {
                yl1 = static_cast <int> (ypos); // floor
                yl2 = static_cast <int> (ypos + 1.0); // ceil
                /*if ((ypos + 1.0) == yl2)
                    --yl2;*/
                yf1 = yl2 - ypos;
                yf2 = ypos - yl1;

                ysrc1 = src + srcrowlen * yl1;
                ysrc2 = src + srcrowlen * yl2;
                xpos = 0.0;
                xdst = dst;

                for (unsigned x = 0; x < width; ++x)
                {
                    xl1 = static_cast <int> (xpos); // floor
                    xl2 = static_cast <int> (xpos + 1.0); // ceil
                    /*if ((xpos + 1.0) == xl2)
                        --xl2;*/
                    xf1 = xl2 - xpos;
                    xf2 = xpos - xl1;

                    *xdst = averageGS  (*(ysrc1 + xl1),
                                        *(ysrc1 + xl2),
                                        *(ysrc2 + xl1),
                                        *(ysrc2 + xl2),
                                        xf1 * yf1,
                                        xf2 * yf1,
                                        xf1 * yf2,
                                        xf2 * yf2);

                    xpos += xstep;
                    ++xdst;
                }

                ypos += ystep;
                dst += dstrowlen;
            }
            break;
        }
    }
    else
        switch (resize)
        {
        case reBilinear:
            ypos = 0.0;
            for (unsigned y = 0; y < height; ++y)
            {
                yl1 = static_cast <int> (ypos); // floor
                yl2 = static_cast <int> (ypos + 1.0); // ceil
                /*if ((ypos + 1.0) == yl2)
                    --yl2;*/
                yf1 = yl2 - ypos;
                yf2 = ypos - yl1;

                ysrc1 = src + srcrowlen * yl1;
                ysrc2 = src + srcrowlen * yl2;
                xpos = 0.0;
                xdst = dst;

                for (unsigned x = 0; x < width; ++x)
                {
                    xl1 = static_cast <int> (xpos); // floor
                    xl2 = static_cast <int> (xpos + 1.0); // ceil
                    /*if ((xpos + 1.0) == xl2)
                        --xl2;*/
                    xf1 = xl2 - xpos;
                    xf2 = xpos - xl1;

                    *reinterpret_cast <DWORD*> (xdst)
                        = averageRGB (*reinterpret_cast <const DWORD*> (ysrc1 + xl1 * 3),
                                      *reinterpret_cast <const DWORD*> (ysrc1 + xl2 * 3),
                                      *reinterpret_cast <const DWORD*> (ysrc2 + xl1 * 3),
                                      *reinterpret_cast <const DWORD*> (ysrc2 + xl2 * 3),
                                      xf1 * yf1,
                                      xf2 * yf1,
                                      xf1 * yf2,
                                      xf2 * yf2);

                    xpos += xstep;
                    xdst += 3;
                }

                ypos += ystep;
                dst += dstrowlen;
            }
            break;
        }

    obmdata = nbmdata;
    obminfo->bmiHeader.biWidth = width;
    obminfo->bmiHeader.biHeight = height;
    obminfo->bmiHeader.biSizeImage = dstrowlen * height;
}

void GSBitmap::performGrayscaleConversion (colortable_t* ctable,
        const colortograyscalefactor_t& _pctgf, const contrast_t& _pcontrast,
        dithering_t dithering)
{
    unsigned colors;
    colortograyscalefactor_t ctgf = _pctgf;
    const unsigned char* src;
    const unsigned char* xsrc;
    unsigned char*       dst;
    ucl::stl::ArrayPtr <BITMAPINFO> gbminfo;
    ucl::stl::ArrayPtr <BYTE>       gbmdata;

    float contrast = bound <0, 1000> (_pcontrast.contrast) / 100.0f;
    float brightness = bound <-100, 100> (_pcontrast.brightness);

    unsigned srcrowlen, dstrowlen;

    unsigned xs, ys;


    if (!obmdata.get ())
        return;
    if (obminfo->bmiHeader.biBitCount == 8) // already grayscaled
        return;

    colors = ctable->colors;
    
    xs = obminfo->bmiHeader.biWidth;
    ys = obminfo->bmiHeader.biHeight;

    srcrowlen = ((xs * 24 + 31) & ~31) >> 3;
    dstrowlen = ((xs * 8 + 31) & ~31) >> 3;

    gbminfo.reset (reinterpret_cast <BITMAPINFO*> (new char[offsetof (BITMAPINFO, bmiColors)
        + sizeof (RGBQUAD) * colors]));

    gbminfo->bmiHeader.biSize = sizeof (gbminfo->bmiHeader);
    gbminfo->bmiHeader.biWidth = xs;
    gbminfo->bmiHeader.biHeight = ys;
    gbminfo->bmiHeader.biPlanes = 1;
    gbminfo->bmiHeader.biBitCount = 8;
    gbminfo->bmiHeader.biCompression = 0;
    gbminfo->bmiHeader.biSizeImage = dstrowlen * ys;
    gbminfo->bmiHeader.biXPelsPerMeter = gbminfo->bmiHeader.biYPelsPerMeter = 0;
    gbminfo->bmiHeader.biClrUsed = colors;
    gbminfo->bmiHeader.biClrImportant = colors;

    gbmdata.alloc (dstrowlen * ys);

    ctgf.normalize ();

    for (unsigned i = 0; i < colors; ++i)
        gbminfo->bmiColors[i] = ctable->rgbvals[i];

    src = obmdata.get ();
    dst = gbmdata.get ();


    typedef float                       lpix_t;
    typedef ucl::stl::ArrayPtr <lpix_t> lrow_t;
    typedef ucl::stl::ArrayPtr <lrow_t> lpic_t;

    float oldpix, quanterr;
    unsigned newpix;
    
    lpic_t pic;
    lrow_t* row0;
    lrow_t* row1;
    lpix_t* pix;
    lpix_t* pmax;

    switch (dithering)
    {
    case diNone:
        for (unsigned y = 0; y < ys; ++y)
        {
            xsrc = src;
            for (unsigned x = 0; x < xs; ++x)
            {
                dst[x] = ctable->findNearestColor (ctgf.getGrayscaleFromRGB (
                    *reinterpret_cast <const DWORD*> (xsrc))
                        * contrast + brightness);
                xsrc += 3;
            }
            src += srcrowlen;
            dst += dstrowlen;
        }
        break;
    case diFloydSteinberg:
        pic.alloc (2);
        row0 = &pic[0];
        row1 = &pic[1];
        row0->alloc (xs + 1);
        row1->alloc (xs + 1);

        (*row0)[0] = 0.0f;
        (*row1)[0] = 0.0f;
        xsrc = src;
        for (unsigned x = 0; x < xs; ++x)
        {
            oldpix = ctgf.getGrayscaleFromRGB (*reinterpret_cast <const DWORD*> (xsrc))
                   * contrast + brightness
                   + (*row0)[x]; // no bounding
            newpix = ctable->findNearestColor (oldpix); // ctbl index
            quanterr = oldpix - ctable->rgbvals[newpix].rgbBlue;

            xsrc += 3;

            dst[x] = newpix;
            (*row0)[x + 1]  = quanterr * (7.0f / 16.0f); // no += since the row is not initialized
            (*row1)[x + 1]  = quanterr * (1.0f / 16.0f);
            (*row1)[x]     += quanterr * (5.0f / 16.0f);
            if (x == 0)
                continue;
            (*row1)[x - 1] += quanterr * (3.0f / 16.0f);
        }


        for (unsigned y = 1; y < ys; ++y)
        {
            src += srcrowlen;
            dst += dstrowlen;
            std::swap (row0, row1);
            (*row1)[0] = 0.0f;

            xsrc = src;
            for (unsigned x = 0; x < xs; ++x)
            {
                oldpix = ctgf.getGrayscaleFromRGB (*reinterpret_cast <const DWORD*> (xsrc))
                       * contrast + brightness
                       + (*row0)[x]; // no bounding
                newpix = ctable->findNearestColor (oldpix); // ctbl index
                quanterr = oldpix - ctable->rgbvals[newpix].rgbBlue;

                xsrc += 3;

                dst[x] = newpix;
                (*row0)[x + 1] += quanterr * (7.0f / 16.0f);
                (*row1)[x + 1]  = quanterr * (1.0f / 16.0f);
                (*row1)[x]     += quanterr * (5.0f / 16.0f);
                if (x == 0)
                    continue;
                (*row1)[x - 1] += quanterr * (3.0f / 16.0f);
            }
        }
        break;
    }

    obmdata = gbmdata;
    obminfo = gbminfo;
}


void GSBitmap::_internal_checkfile_size (const BITMAPINFOHEADER* ptr,
    int Width, int Height)
{
    char buf[256];
    if ((ptr->biWidth != Width) && (ptr->biHeight != Height))
    {
        std::sprintf (buf, "Wrong size (required: %dx%d, given: %dx%d)",
            Width, Height, ptr->biWidth, ptr->biHeight);
        throw std::runtime_error (buf);
    }
}

void GSBitmap::_internal_checkfile_rel (const BITMAPINFOHEADER* ptr, int X, int Y)
{
    char buf[256];
    ucl::math::Fraction req (X, Y), giv (ptr->biWidth, ptr->biHeight);

    if (req != giv)
    {
        std::sprintf (buf, "Wrong size relation (required: %d:%d, given: %d:%d)",
            req.getNumerator (), req.getDenominator (),
            giv.getNumerator (), giv.getDenominator ());
        throw std::runtime_error (buf);
    }
}

void GSBitmap::drawToDevice (HDC hdc, unsigned x, unsigned y)
{
    const BYTE* bmd = getBitmap ();
    const BITMAPINFO* bmi = getBitmapInfo ();

    if (!(bmd && bmi))
        return;

    SetDIBitsToDevice (hdc, x, y, bmi->bmiHeader.biWidth,
        bmi->bmiHeader.biHeight, 0, 0, 0, bmi->bmiHeader.biHeight,
        bmd, bmi, DIB_RGB_COLORS);
}


} // namespace agraph

