
/*
 *  gsconvert.hpp
 *
 *  A class that supports several bitmap operations:
 *  - loading and saving bitmaps
 *  - downsize bitmaps
 *  - convert bitmaps to grayscale pictures
 *
 *  Written by Moritz Beutel
 *  (c) 2007 by AUDACIA Software
 */

#ifndef _GSCONVERT_HPP
#define _GSCONVERT_HPP

#include <ucl/stl/arrayptr.hpp>
#include <windows.h>


namespace agraph
{

struct colortograyscalefactor_t
{
    float blue_factor;
    float green_factor;
    float red_factor;

    colortograyscalefactor_t (void)
    { setToDefault (); }
    
    void normalize (void);
    colortograyscalefactor_t& setToDefault (void);
    inline float getGrayscaleFromRGB (unsigned rgb) const
    {
        float val =
               (rgb & 0xFF)         * blue_factor
             + ((rgb >> 8) & 0xFF)  * green_factor
             + ((rgb >> 16) & 0xFF) * red_factor
             ;
        return val;
    }

};

struct colortable_t
{
private:
    colortable_t (unsigned clrs)
     : colors (clrs)
    {}

public:
    const unsigned  colors;
    RGBQUAD         rgbvals[0];

    static ucl::stl::ArrayPtr <colortable_t> createColorTable (unsigned char colors);
    colortable_t& setToDefault (void);
    unsigned findNearestColor (float rgb) const;
    unsigned findNearestColor (unsigned char rgb) const;
};


class GSBitmap
{
private:
    static void _internal_checkfile_size (const BITMAPINFOHEADER* ptr,
        int Width, int Height);
    static void _internal_checkfile_rel (const BITMAPINFOHEADER* ptr,
        int X, int Y);


public:
    struct contrast_t
    {
        int contrast;
        int brightness;

        contrast_t (void)
         : contrast (100), brightness (0)
        {}
        contrast_t (int contr, int bright)
         : contrast (contr), brightness (bright)
        {}
    };

    enum dithering_t
    {
        diNone,
        diFloydSteinberg,
    };

    enum resize_t
    {
        reBilinear,
    };

    typedef void (*checkfile_t) (void*, const BITMAPINFOHEADER*);

    template <int Width, int Height>
        static void checkfile_size (void*, const BITMAPINFOHEADER* ptr)
    { _internal_checkfile_size (ptr, Width, Height); }
    template <int X, int Y>
        static void checkfile_relation (void*, const BITMAPINFOHEADER* ptr)
    { _internal_checkfile_rel (ptr, X, Y); }


private:
    ucl::stl::ArrayPtr <colortable_t> ctbl;
    colortograyscalefactor_t          cgsf;

    ucl::stl::ArrayPtr <BITMAPINFO>   obminfo;
    ucl::stl::ArrayPtr <BYTE>         obmdata;


public:
    GSBitmap (void);
    GSBitmap (const GSBitmap& rhs);
    ~GSBitmap (void);

    GSBitmap& operator = (const GSBitmap& rhs);

    const BYTE* getBitmap (void)
    { return obmdata.get (); }
    const BITMAPINFO* getBitmapInfo (void)
    { return obminfo.get (); }

    void loadFromBitmap (const char* filename, checkfile_t cf = 0, void* cfdata = 0);
    void pasteFromBitmap (const BITMAPINFOHEADER* bmi, const BYTE* bdata,
        checkfile_t cf = 0, void* cfdata = 0);
    void saveToBitmap (const char* filename);

    void performResize (unsigned width, unsigned height, resize_t resize);

    void performGrayscaleConversion (colortable_t* ctable,
        const colortograyscalefactor_t& ctgf = colortograyscalefactor_t (),
        const contrast_t& contrast = contrast_t (),
        dithering_t dithering = diFloydSteinberg);

    void drawToDevice (HDC hdc, unsigned x = 0, unsigned y = 0);
};


} // namespace apv


#endif // _AGSFILE_HPP
