
/*
 *  ucl/_debug.hpp
 *
 *  Debugging support definitions
 *  Author: Moritz Beutel
 */


#ifndef _UCL_DEBUG_HPP
#define _UCL_DEBUG_HPP


namespace ucl
{

/**
 *  @brief   Various debugging definitions
 *
 *  Define the following macros to enable or disable debugging:
 *
 *  - <b>_DEBUG</b> - All debugging options enabled
 *
 *  - <b>UCL_DEBUG</b> - All debugging options enabled
 *  - <b>UCL_NO_DEBUG</b> - All debugging options disabled
 *
 *  - <b>UCL_ITERATOR_DEBUGGING</b> - Iterator debugging enabled
 *  - <b>UCL_NO_ITERATOR_DEBUGGING</b> - Iterator debugging disabled
 *
 *  - <b>UCL_DEBUG_EXCEPTION_STD</b> - std::exception derivate thrown if debug
 *    assertion occurs (default)
 *  - <b>UCL_DEBUG_EXCEPTION_CERR</b> - If debug assertion occurs, it is logged
 *    with std::cerr.
 *    @warning If this symbol is defined, iterator debugging assertions do not
 *             interrupt the program, therefore it may behave undefinedly.
 *  - <b>UCL_DEBUG_EXCEPTION_DELPHI</b> - Throws Sysutils::Exception derivate
 *    if debug assertion occurs. This might be necessary in VCL applications to
 *    assure exceptions are catched and shown properly by the VCL
 *    (C++Builder only)
 *  - <b>UCL_DEBUG_EXCEPTION_OUTPUTDEBUGSTRING</b> - If debug assertion occurs,
 *    it is logged with OutputDebugString (Win32 only)
 *    @warning If this symbol is defined, iterator debugging assertions do not
 *             interrupt the program, therefore it may behave undefinedly.
 */

namespace debug
{

} // namespace debug

} // namespace ucl


    // first, define everything
#if defined(_DEBUG) && !defined(UCL_DEBUG)
 #define UCL_DEBUG
 #define UCL_ITERATOR_DEBUGGING
 #define UCL_DEBUG_EXCEPTION_STD
#endif
#if defined(UCL_NO_DEBUG)
 #undef UCL_DEBUG
#endif

    // now undef everything that is explicitly disabled
#if defined(UCL_NO_ITERATOR_DEBUGGING)
 #undef UCL_ITERATOR_DEBUGGING
#endif
#if defined(UCL_DEBUG_EXCEPTION_DELPHI)            \
 || defined(UCL_DEBUG_EXCEPTION_OUTPUTDEBUGSTRING) \
 || defined(UCL_DEBUG_EXCEPTION_CERR)
 #undef UCL_DEBUG_EXCEPTION_STD
#endif



#ifdef UCL_DEBUG
 #include <string>
#endif

#ifdef UCL_ITERATOR_DEBUGGING
 #include <typeinfo>

 #define _UCL_ITERWRAP_CONSTRUCT    , this
#endif

#ifdef UCL_DEBUG_EXCEPTION_STD
 #include <stdexcept>
#endif
#ifdef UCL_DEBUG_EXCEPTION_CERR
 #include <iostream>
#endif
#ifdef UCL_DEBUG_EXCEPTION_DELPHI
 #include <SysUtils.hpp>
#endif
#ifdef UCL_DEBUG_EXCEPTION_OUTPUTDEBUGSTRING
 #include <windows.h>
#endif


#if defined(UCL_DEBUG_EXCEPTION_CERR)
 #define _UCL_DEBUG_THROW(type,msg) \
    std::cerr << #type ": " << msg << std::endl
#elif defined(UCL_DEBUG_EXCEPTION_DELPHI)
 #define _UCL_DEBUG_THROW(type,msg) \
    throw Sysutils::Exception ((std::string (#type ": ") + msg).c_str ())
#elif defined(UCL_DEBUG_EXCEPTION_OUTPUTDEBUGSTRING)
 #define _UCL_DEBUG_THROW(type,msg) \
    OutputDebugString ((std::string (#type ": ") + msg).c_str ())
#else
 #define _UCL_DEBUG_THROW(type,msg) \
    throw type (msg)
#endif

#ifdef UCL_DEBUG
 #define _UCL_DEBUG_STATEMENT(s)        s
 #define _UCL_THROW(type,msg) _UCL_DEBUG_THROW(type,msg)
 #define _UCL_ASSERT(cond,type,msg) if (!(cond)) _UCL_DEBUG_THROW(type,msg)
#else
 #define _UCL_DEBUG_STATEMENT(s)
 #define _UCL_THROW(type,msg)
 #define _UCL_ASSERT(cond,type,msg)
#endif

#ifdef UCL_ITERATOR_DEBUGGING
 #define _UCL_ITERDBG_STATEMENT(s)    _UCL_DEBUG_STATEMENT(s)
 #define _UCL_ITERDBG_THROW(type,msg) _UCL_THROW(type,msg)
#else
 #define _UCL_ITERDBG_STATEMENT(s)
 #define _UCL_ITERDBG_THROW(type,msg)
#endif

#endif // _UCL_DEBUG_HPP
