
/*
 *  ucl/bcc/multicast.hpp
 *
 *  Multicast Closures for C++Builder
 *  Author: Moritz Beutel
 *
 *  The MulticastClosureN class template and the asMulticast()
 *  function template provide an easy use of multicast closures for
 *  Borland/CodeGear C++Builder.
 */

#ifndef _UCL_MULTICAST_HPP
#define _UCL_MULTICAST_HPP

#include <ucl/bcc/multicastsupport.hpp>


#include <list>
#include <cstddef>

#include <ucl/stl/iterwrap.hpp>
#include <ucl/stl/arrayptr.hpp>


namespace ucl
{

namespace bcc
{


namespace detail
{

/// Non-template base class that implements most functionality of
/// MulticastClosure. Please refer to the documentation for
/// MulticastClosureN for more information.
class MulticastClosureBase
{
    friend struct MulticastContainer;

protected:
    static MulticastClosureBase* registerMulticastClosure (/*bool&, MulticastFactory,*/
        MulticastContainer&, TMethod*, const TMethod*, MulticastClosureBase*);

    static const unsigned MCEBSignature;

    typedef std::list <TMethod> list_t;


         /// @brief Signature pattern to differentiate between MulticastClosure
         ///        objects and event handler classes
    const unsigned signature;

         /// @brief More controllable "replacement" for VTable pointer
    MulticastDestructor mcd;

    MulticastClosureBase* next;

    list_t list;

    TMethod* myEvent;


    MulticastClosureBase (MulticastDestructor md, TMethod* event);
    ~MulticastClosureBase (void);

    void _push_back (const void* e);
    void _push_front (const void* e);
    void _remove (const void* e);
    bool _contains (const void* e);
    void _swap (MulticastClosureBase& rhs);
    list_t::iterator _insert (list_t::iterator i, const void* val);
    list_t::iterator _erase (list_t::iterator i);

         /// @brief Destroys this MulticastClosure and calls .destroy() for the
         ///        linked MulticastClosure object.
    void destroy (void) { mcd (this); }

public:
        /// Returns the amount of event handlers referenced by the MulticastClosure
    std::size_t size (void) { return list.size (); }
        /// Determines whether the event list is empty
    bool empty (void) { return (list.empty ()); }
        /// Determines whether the event list is empty
    void clear (void);
        /// Removes the last event in the list
    void pop_back (void);
        /// Removes the first event in the list
    void pop_front (void);

        /// @brief Compares the signature to a special pattern value to verify
        ///        that the current object is a MulticastClosure
    bool isMulticastClosure (void) { return (this ? (signature == MCEBSignature) : false); }
};


/// Simple template base class that defines most of the container
/// interfaces MulticastClosure. Please refer to the documentation for
/// MulticastClosureN for more information.
template <typename ClosureT>
    class MulticastClosureCont : public MulticastClosureBase
{
    typedef MulticastClosureCont <ClosureT> Base;

protected:
    MulticastClosureCont (MulticastDestructor mf, TMethod* event)
     : MulticastClosureBase (mf, event) {}

public:
    typedef stl::BiDiIterator <ClosureT, Base, UCL_ITERWRAP_FORWARD,
        list_t::iterator> iterator;
    typedef stl::BiDiIterator <const ClosureT, Base, UCL_ITERWRAP_FORWARD,
        list_t::const_iterator> const_iterator;
    typedef stl::BiDiIterator <ClosureT, Base, UCL_ITERWRAP_REVERSE,
        list_t::reverse_iterator> reverse_iterator;
    typedef stl::BiDiIterator <const ClosureT, Base, UCL_ITERWRAP_REVERSE,
        list_t::const_reverse_iterator> const_reverse_iterator;

        /// Returns an iterator that points to the first event in the list
    iterator begin (void)
    { return iterator (list.begin () _UCL_ITERWRAP_CONSTRUCT); }
        /// Returns an const_iterator that points to the first event in the list
    const_iterator begin (void) const
    { return const_iterator (list.begin () _UCL_ITERWRAP_CONSTRUCT); }
        /// Returns an iterator that points beyond the last event in the list
    iterator end (void)
    { return iterator (list.end () _UCL_ITERWRAP_CONSTRUCT); }
        /// Returns an const_iterator that points beyond the last event in the list
    const_iterator end (void) const
    { return const_iterator (list.end () _UCL_ITERWRAP_CONSTRUCT); }

        /// Returns an reverse_iterator that points to the last event in the list
    reverse_iterator rbegin (void)
    { return reverse_iterator (list.rbegin () _UCL_ITERWRAP_CONSTRUCT); }
        /// Returns an const_reverse_iterator that points to the last event in the list
    const_reverse_iterator rbegin (void) const
    { return const_reverse_iterator (list.rbegin () _UCL_ITERWRAP_CONSTRUCT); }
        /// Returns an reverse_iterator that points beyond the first event in the list
    reverse_iterator rend (void)
    { return reverse_iterator (list.rend () _UCL_ITERWRAP_CONSTRUCT); }
        /// Returns an const_reverse_iterator that points beyond the first event in the list
    const_reverse_iterator rend (void) const
    { return const_reverse_iterator (list.rend () _UCL_ITERWRAP_CONSTRUCT); }

        /// Returns a reference to the last event in the list
    ClosureT& back (void) { return *reinterpret_cast <ClosureT*> (&list.back ()); }
        /// Returns a const_reference to the last event in the list
    const ClosureT& back (void) const { return *reinterpret_cast <const ClosureT*> (&list.back ()); }
        /// Returns a reference to the first event in the list
    ClosureT& front (void) { return *reinterpret_cast <ClosureT*> (&list.front ()); }
        /// Returns a const_reference to the first event in the list
    const ClosureT& front (void) const { return *reinterpret_cast <const ClosureT*> (&list.front ()); }

        /// Inserts an event before the given iterator
    iterator insert (iterator where, const ClosureT& val)
    { return iterator (_insert (where._getWrappedIterator (), &val) _UCL_ITERWRAP_CONSTRUCT); }
        /// Erases the event referenced by the given iterator
    iterator erase (iterator where)
    { return iterator (_erase (where._getWrappedIterator ()) _UCL_ITERWRAP_CONSTRUCT); }

        /// Attaches an event to the end of the list
    void push_back (const ClosureT& val)
    { _push_back (&val); }
        /// Attaches an event to the begin of the list
    void push_front (const ClosureT& val)
    { _push_front (&val); }
        /// Removes all occurances of a specific event
    void remove (const ClosureT& val)
    { _remove (&val); }
        /// Determines whether the MulticastClosure contains a specific event
    bool contains (const ClosureT& val)
    { return _contains (&val); }
        /// Swaps the events of two MulticastClosure objects
    void swap (Base& rhs)
    { _swap (rhs); }
};

} // namespace detail


#ifdef _DOXYGEN

// doxygen friendly version

/**
 *  @brief   Closure container that can be used as closure destination
 *  @author  Moritz Beutel
 *
 *  This class provides access to a list of event handlers. The operator ()
 *  and the dispatch() method can be used as an event handler for closures
 *  to make them capable of handling multicasts (this is what asMulticast() does).
 *  The class provides a container interface and iterator and reverse_iterator
 *  types for easily iterating over the event list.
 *
 *  The template parameters are:
 *  - <b>typename Retval</b> - return value of the closure
 *  - <b>typename Param1, ..., typename ParamN</b> - the parameter types of the closure
 *  - <b>typename CC = ccFastcall</b> - a calling convention descriptor.
 *    possible values are:
 *      -# <b>ccCdecl</b> - __cdecl (default) calling convention
 *      -# <b>ccPascal</b> - __pascal calling convention
 *      -# <b>ccStdcall</b> - __stdcall calling convention
 *      -# <b>ccFastcall</b> - __fastcall calling convention
 *
 *  To instantiate, use the identifier 'MulticastClosure' with a numeric suffix
 *  that represents the number of function parameters, e.g.:
 *
 *  <c>ucl::MulticastClosure1 <void, TObject*>& mc = asMulticast <void, TObject*> (Button1->OnClick);</c> 
 */

template <typename Retval,
          typename Param1, ..., typename ParamN,
          typename CC = ccFastcall>
    class MulticastClosureN
        : public detail::MulticastClosureCont
            <typename detail::FunctorTypeN <CC, Retval, Param1, ..., ParamN>::Closure>
{
public:
        /// Typedef for the closure type
    typedef typename Retval __calling_convention (__closure* ClosureT) (Param1, ..., ParamN);

private:
    typedef detail::MulticastClosureCont <ClosureT> ContBase;

public:
        /// MulticastClosure Factory function
    static detail::MulticastClosureBase* getMulticastClosure (void);

        /**
         *  @brief MulticastClosure destructor replacement
         *
         *  As described at isMulticast(), a MulticastClosure is recognized
         *  by a signature pattern in the first dword.
         *  Of course, this requires that MulticastClosureBase does not
         *  have a VTable pointer and therefore is not polymorphic. For
         *  that reason, the destructor is not virtual but the class
         *  contains a function pointer that points to this destructor
         *  function.
         */
    static void deleteMulticastClosure (detail::MulticastClosureBase* ptr);

    MulticastClosureN (void) {}
    ~MulticastClosureN (void) {}

        /**
         *  @brief Default MulticastClosure dispatcher
         *
         *  This method is called when the event is triggered. It iterates
         *  through the event list and calls every registered event with the
         *  given parameters. The return values are discarded.
         *
         *  This function does not catch EMulticastBreak exceptions. Use the
         *  dispatch() method for that purpose.
         */
    Retval __calling_convention operator () (Param1 p1, ..., ParamN pn)
    {
        list_t::iterator i, end;

        for (i = list.begin (), end = list.end (); i != end; ++i)
            if (i->Code)
                (*reinterpret_cast <ClosureT*> (&*i)) (p1, ..., pn);

        return Retval (); // return a default-constructed retval
    }

        /**
         *  @brief Interruptable MulticastClosure dispatcher
         *
         *  If the MulticastClosure object is terminable, this method is
         *  called when the event is triggered. It iterates through the
         *  event list and calls every registered event with the given
         *  parameters. The return values are discarded.
         *
         *  This function catches EMulticastBreak exceptions. It can be set
         *  as event handler function with setTerminable().
         */
    Retval __calling_convention dispatch (Param1 p1, ..., ParamN pn)
    {
        try
        { (*this) (p1, ..., pn); }
        catch (EMulticastBreak&)
        { }

        return Retval (); // return a default-constructed retval
    }

        /**
         *  @brief MulticastClosure dispatcher with return value management
         *
         *  This method iterates through the event list and calls every
         *  registered event with the given parameters. The return values
         *  are written to an array of Retvals and returned as ArrayPtr <Retval>.
         *
         *  This function catches EMulticastBreak exceptions.
         */
    ArrayPtr <Retval> __calling_convention dispatchAndGetResults
        (Param1 p1, ..., ParamN pn)
    {
        list_t::iterator i, end;
        Retval* res_base = new Retval[list.size ()];
        Retval* res = res_base;

        try
        {
            for (i = list.begin (), end = list.end (); i != end; ++i)
            {
                if (i->Code)
                    *res = ((*reinterpret_cast <ClosureT*> (&*i)) (p1, ..., pn));
                ++res;
            }
        }
        catch (EMulticastBreak&)
        { }

        return ArrayPtr <Retval> (res);
    }


        /**
         *  @brief   Returns the MulticastClosure object associated with the
         *           given closure
         *
         *  This function returns a reference to the MulticastClosure object
         *  associated with the closure that was given as parameter. If the
         *  closure does not refer to a MulticastClosure object (please read
         *  the documentation of the isMulticast() function for more information
         *  about this), the function creates one and adjusts the closure
         *  (the current event handler is added to the MulticastClosure).
         *
         *  @warning Do not call this function directly. Call the global
         *  asMulticast() function instead.
         */
    static MulticastClosureN* _asMulticast (MulticastContainer& mcc, ClosureT& e)
    {
        TMethod* meth = reinterpret_cast <TMethod*> (&e);
        MulticastClosureN* data
            = reinterpret_cast <MulticastClosureN*> (meth->Data);

        if (data && (*reinterpret_cast <unsigned*> (data) == MCEBSignature))
        {
            data->myEvent = meth;
            return data;
        }
        else
        {
            data = new MulticastClosureN (meth);
            ClosureT be;
            be = data->operator ();
            return reinterpret_cast <MulticastClosureN*>
                (registerMulticastClosure (mcc, meth,
                    reinterpret_cast <TMethod*> (&be), data));
        }
    }

        /** @brief   Enables or disables the support for aborting the execution
         *           of a MulticastClosure event
         *
         *  To break the execution of an event, throw the ucl::bcc::EMulticastBreak
         *  exception in a event handler.
         */
    void setTerminable (bool terminable)
    {
        ClosureT cl;

        if (reinterpret_cast <TMethod*> (&cl)->Data)
        {
            if (terminable)
                cl = this->dispatch;
            else
                cl = this->operator ();
            *reinterpret_cast <ClosureT*> (myEvent) = cl;
        }
    }

        /** @brief   Determines whether a MulticastClosure event is terminable
         *
         *  To terminate the execution of an event, throw the ucl::bcc::EMulticastBreak
         *  exception in a event handler.
         */
    bool getTerminable (void)
    {
        ClosureT cl;
        cl = this->dispatch;
        return (*reinterpret_cast <ClosureT*> (myEvent) == cl);
    }
};

/**
 *  @brief   Returns the MulticastClosure object associated with the
 *           given closure
 *  @author  Moritz Beutel
 *
 *  This function returns a reference to the MulticastClosure object
 *  associated with the closure that was given as parameter. If the
 *  closure does not refer to a MulticastClosure object (please read
 *  the documentation of the isMulticast() function for more information
 *  about this), the function creates one and adjusts the closure
 *  (the current event handler is added to the MulticastClosure).
 */
 template <typename Retval, typename Param1, ..., ParamN>
    MulticastClosureN <Retval, Param1, ..., ParamN>& asMulticast
    (MulticastContainer& mcc, Retval __calling_convention (__closure*& e) (Param1, ...))
{ return MulticastClosureN <Retval, Param1, ..., ParamN>::_asMulticast (mcc, e); }


#else

// the real definition

#include <ucl/bcc/_multicastdef.hpp>

#endif

} // namespace bcc

} // namespace ucl


#endif // _UCL_MULTICAST_HPP
