
#ifndef _UCL_MULTICASTSUPPORT_HPP
#define _UCL_MULTICASTSUPPORT_HPP


#ifndef __BCPLUSPLUS__
 #error multicast.hpp implements an extension of the __closure keyword\
 that is specific to Borland C++. Do not attempt to use it with other compilers.
#endif


#include <Sysutils.hpp>
#ifndef UCL_DEBUG_EXCEPTION_DELPHI
 #include <stdexcept>
#endif

    /*
     * define the UCL_DEBUG_EXCEPTION_DELPHI macro to derive the
     * EMulticastAbort class from VCL's Sysutils::Exception class
     */


namespace ucl
{

namespace bcc
{

struct ccCdecl;
struct ccPascal;
struct ccStdcall;
struct ccFastcall;


class MulticastContainer;


    /// @cond UCL_DETAIL
namespace detail
{

class MulticastClosureBase;

typedef void (*MulticastDestructor) (MulticastClosureBase*);

} // namespace detail
    /// @endcond


/**
 *  @brief   Manages and destroys MulticastClosure objects
 *  @author  Moritz Beutel
 *
 *  This class is considered to be the owner of MulticastClosure objects.
 *  It contains a pointer to the first and the last MulticastClosure object
 *  (the objects are linked together) to ease attaching and destroying.
 *
 *  Put an instance of this class into the class that contains the event
 *  handlers you will link to.
 *  @warning The MulticastClosure objects referenced by this class will
 *           be destroyed when the MulticastContainer is destroyed!
 */

class MulticastContainer
{
    friend class detail::MulticastClosureBase;

private:
    detail::MulticastClosureBase* first;
    detail::MulticastClosureBase* last;

    MulticastContainer (const MulticastContainer&);
    const MulticastContainer& operator = (const MulticastContainer&);

public:
    MulticastContainer (void) : first (NULL), last (NULL) { }
    ~MulticastContainer (void);
};


/**
 *  @brief   Exception class for aborting the dispatch of a MulticastClosure call
 *  @author  Moritz Beutel
 *
 *  Throw an object of this class in your event handler if you want to abort
 *  the dispatch of a MulticastClosure call. This exception is not caught by
 *  the operator () of MulticastClosures; use the dispatch() member function
 *  to enable interrupting (or catch the exception yourself).
 *  You can enable exception catching for existing events by calling the
 *  setTerminable() method.
 *
 *  This class derives from std::exception per default. To make it derive from
 *  Sysutils::Exception, define the preprocessor symbol
 *  UCL_DEBUG_EXCEPTION_DELPHI.
 */

#ifdef UCL_DEBUG_EXCEPTION_DELPHI
class EMulticastBreak : public Exception
{
public:
    EMulticastBreak (void) : Sysutils::Exception ("Multicast aborted") {}
};
#else
class EMulticastBreak : public std::exception
{
public:
    virtual const char* what() const
    { return "Multicast aborted"; }
};
#endif


/**
 *  @brief   Determines whether a closure is a MulticastClosure
 *  @author  Moritz Beutel
 *
 *  This function determines whether a closure is a MulticastClosure.
 *  Unlike asMulticast(), it does not touch the closure itself.
 *
 *  The function finds out about that by comparing the first dword
 *  of the object referenced by the closure data pointer to 0xFEDCBA98.
 *  The comparison of the first dword of the object to an object
 *  signature is a dirty trick to determine whether the closure refers
 *  to a MulticastClosure. Classes derived from TObject store
 *  the VTable pointer in the first dword. A VTable pointer
 *  always points to an address in the user space, so it must be
 *  below 0x80000000 (in usual applications) or 0xC0000000 (for
 *  3 GB enabled applications). Therefore, it works well for VCL
 *  classes. However, if you use multicast closures in non-VCL classes,
 *  make sure that either all closures are MulticastClosures or your
 *  event handler class has something in the first dword that can
 *  impossibly be 0xFEDCBA98, e.g. a user space pointer. 
 */

template <typename ClosureT>
    bool isMulticast (const ClosureT& e)
{
    return reinterpret_cast <detail::MulticastClosureBase*>
        (reinterpret_cast <const TMethod*> (&e)->Data)->isMulticastClosure ();
}


} // namespace bcc

} // namespace ucl


#endif // _UCL_MULTICASTSUPPORT_HPP
