
#pragma hdrstop

#include <ucl/win/thread.hpp>
#include <stdexcept>
#include <typeinfo>
#include <process.h>


#include <ucl/_debug.hpp>
#ifdef UCL_EXCEPTION_DELPHI
 #include <SysUtils.hpp>
#endif


namespace ucl
{

namespace win
{

DWORD WINAPI BasicThread::ThreadFunc (LPVOID data)
{
    BasicThread* bt = (BasicThread*) data;

    bt->execute ();
    
    _endthread ();
    return 0;
}


BasicThread::BasicThread (SECURITY_ATTRIBUTES* sa)
 : running (false), cancelled (false)
{
    hThread = (HANDLE) (_beginthreadNT ((void (*)(void*)) ThreadFunc,
        0, this, sa, CREATE_SUSPENDED, (unsigned long*) (&dwThreadId)));
}
BasicThread::~BasicThread (void)
{
    CloseHandle (hThread);
}

void BasicThread::suspend (void)
{
    if (running)
    {
        running = false;
        SuspendThread (hThread);
        running = true;
    }
}
void BasicThread::resume (void)
{
    if (hThread && !running)
    {
        running = true;
        ResumeThread (hThread);
    }
}
void BasicThread::cancel (void)
{
    cancelled = true;
}
void BasicThread::terminate (void)
{
    if (hThread)
    {
        TerminateThread (hThread, 0);
        CloseHandle (hThread);
        hThread = 0;
        dwThreadId = 0;
        running = false;
    }
}
void BasicThread::reset (void)
{
    if (hThread)
        CloseHandle (hThread);
    hThread = (HANDLE) (_beginthreadNT ((void (*)(void*)) ThreadFunc,
        0, this, 0, 0, (unsigned long*) (&dwThreadId)));
}

DWORD BasicThread::getExitCode (void) const
{
    DWORD r;
    if (GetExitCodeThread (hThread, &r) == STILL_ACTIVE)
        r = STILL_ACTIVE;
    return r;
}



DWORD ThreadPool::WorkThread::execute (void)
{
    bool isNew = true;

    while (true)
    {
            // get task
        EnterCriticalSection (&(owner->cs));

        if (isNew)
        {
            if (owner->threads_active++ == 0)
                (owner->msgfunc) (owner->msgdata, ThreadPool::msgBusy, task);
            isNew = false;
        }

        if (owner->tasks.size () > 0)
        {
            task = owner->tasks.front ();
            owner->tasks.pop_front ();
        }
        else
        {
            if (--owner->threads_active == 0)
                (owner->msgfunc) (owner->msgdata, ThreadPool::msgIdle, task);
            LeaveCriticalSection (&(owner->cs));

            waiting = true;
            WaitForSingleObject (hEvent, INFINITE);
            waiting = false;
            
            if (owner->threads_active++ == 0)
                (owner->msgfunc) (owner->msgdata, ThreadPool::msgBusy, task);

            continue;
        }


        task->threadidx = index;
        task->status = Task::tsBusy;
        
        LeaveCriticalSection (&(owner->cs));

            // perform task
        try
        {
            (owner->msgfunc) (owner->msgdata, ThreadPool::msgTaskBegin, task);
            (task->func) (task->getTaskObj (), *task);
            task->status = Task::tsCompleted;
            (owner->msgfunc) (owner->msgdata, ThreadPool::msgTaskCompleted, task);
        }
        catch (const std::exception& e)
        {
            task->msg = _UCL_DEBUG_STATEMENT (std::string (typeid (e).name ()) + ": " +)
                        e.what ();
            task->status = Task::tsFailed;
            (owner->msgfunc) (owner->msgdata, ThreadPool::msgException, task);
        }
#ifdef UCL_EXCEPTION_DELPHI
        catch (Sysutils::Exception& e)
        {
            task->msg = _UCL_DEBUG_STATEMENT (std::string (e.ClassName ().
                operator AnsiString ().c_str ()) + ": " +) e.Message.c_str ();
            task->status = Task::tsFailed;
            (owner->msgfunc) (owner->msgdata, ThreadPool::msgException, task);
        }
#endif
        catch (...)
        {
            task->msg = "Unknown exception";
            task->status = Task::tsFailed;
            (owner->msgfunc) (owner->msgdata, ThreadPool::msgException, task);
        }

        task.reset (0);
    }
}

ThreadPool::WorkThread::WorkThread (ThreadPool* o, unsigned idx)
 : owner (o), index (idx), hEvent (CreateEvent (NULL, false, false, NULL)),
   waiting (false)
{
}
ThreadPool::WorkThread::~WorkThread (void)
{
    CloseHandle (hEvent);
}


void ThreadPool::_MsgHandler (void* data, unsigned msg, ucl::stl::SmartPtr <Task> task)
{
    msgdata_t* md = reinterpret_cast <msgdata_t*> (data);

    SendMessage (md->hWnd, md->uMsg,
        (WPARAM) msg, (LPARAM) (new ucl::stl::SmartPtr <Task> (task)));
}

void ThreadPool::_init (unsigned thread_cnt)
{
    SYSTEM_INFO si;
    unsigned idx = 0;

    if (thread_cnt == 0)
    {
        GetSystemInfo (&si);
        thread_cnt = si.dwNumberOfProcessors;
    }
    threads.resize (thread_cnt);
    for (std::vector <PWorkThread>::iterator i = threads.begin (), e = threads.end ();
         i != e; ++i, ++idx)
    {
        i->reset (new WorkThread (this, idx));
        (*i)->resume ();
    }

    InitializeCriticalSection (&cs);
}

ThreadPool::ThreadPool (msgfunc_t msg_func, void* msg_data, unsigned thread_cnt)
 : msgfunc (msg_func), msgdata (msg_data), flags (0), threads_active (0)
{
    _init (thread_cnt);
}
ThreadPool::ThreadPool (HWND hWnd, UINT uMsg, unsigned thread_cnt)
 : msgfunc (_MsgHandler), msgdata (new msgdata_t (hWnd, uMsg)),
   flags (0x00000001), threads_active (0)
{
    _init (thread_cnt);
}
ThreadPool::~ThreadPool (void)
{
    //EnterCriticalSection (&cs);   // not necessary - we're terminating everything!
                                    // besides, this could even cause a deadlock.

    if (flags & 0x00000001)
        delete ((msgdata_t*) msgdata);

    for (std::vector <PWorkThread>::iterator i = threads.begin (), e = threads.end ();
         i != e; ++i)
        (*i)->terminate ();

    //LeaveCriticalSection (&cs);
    DeleteCriticalSection (&cs);
}

ucl::stl::SmartPtr <ThreadPool::Task> ThreadPool::addTask (taskfunc_t taskfunc,
    void* obj, void* taskdata)
{
    ucl::stl::SmartPtr <Task> task (new Task (this, taskfunc, obj, taskdata));

    EnterCriticalSection (&cs);

    tasks.push_back (task);
    task->iter = tasks.end ();
    --(task->iter);

    if (threads_active < threads.size ())
        for (std::vector <PWorkThread>::iterator i = threads.begin (), e = threads.end ();
             i != e; ++i)
            if ((*i)->waiting)
            {
                (*i)->waiting = false;
                SetEvent ((*i)->hEvent);
                break;
            }

    LeaveCriticalSection (&cs);

    return task;
}

void ThreadPool::_terminateTask (Task* task)
{
    std::vector <PWorkThread>::iterator i;


    EnterCriticalSection (&cs);

    switch (task->status)
    {
    case Task::tsBusy:
        i = threads.begin () + task->threadidx;
        task->status = Task::tsTerminated;
        if ((*i)->isRunning ())
            --threads_active;
        (*i)->terminate ();
        i->reset (new WorkThread (this, task->threadidx));
        (*i)->resume ();
        break;

    case Task::tsPending:
        tasks.erase (task->iter);
        task->status = Task::tsTerminated;
        break;
    }

    LeaveCriticalSection (&cs);
}

void ThreadPool::_cancelPending (void)
{
    for (std::list <PTask>::iterator i = tasks.begin (), e = tasks.end ();
         i != e; ++i)
    {
        (*i)->status = Task::tsTerminated;
        msgfunc (msgdata, ThreadPool::msgTaskTerminated, *i);
    }
    tasks.clear ();
}
void ThreadPool::terminateAllTasks (void)
{
    unsigned idx;
    

    EnterCriticalSection (&cs);

    for (std::vector <PWorkThread>::iterator i = threads.begin (), e = threads.end ();
         i != e; ++i)
        if ((*i)->isRunning ())
            if ((*i)->task.get ())
            {
                msgfunc (msgdata, ThreadPool::msgTaskTerminated, (*i)->task);
                idx = (*i)->index;
                (*i)->task->status = Task::tsTerminated;
                (*i)->terminate ();
                i->reset (new WorkThread (this, idx));
                (*i)->resume ();
            }
    threads_active = 0;

    _cancelPending ();

    LeaveCriticalSection (&cs);
}
void ThreadPool::cancelPendingTasks (void)
{
    EnterCriticalSection (&cs);
    _cancelPending ();
    LeaveCriticalSection (&cs);
}





} // namespace win

} // namespace ucl

