
/*
 *  ucl/stl/arrayptr.hpp
 *
 *  A std::auto_ptr equivalent that works with heap arrays
 *  Author: Moritz Beutel
 */

#ifndef _UCL_ARRAYPTR_HPP
#define _UCL_ARRAYPTR_HPP

#include <ucl/_debug.hpp>
#include <stdexcept>
#include <cstdlib>
#include <ucl/stl/poddecl.hpp>


namespace ucl
{

namespace stl
{

namespace detail
{

template <unsigned Val>
    struct _ap_copy_t;

template <>
    struct _ap_copy_t <1>
{
    template <typename T>
        static void copy (T* dst, const T* src, unsigned cnt)
    { std::memcpy (dst, src, cnt * sizeof (T)); }
};
template <>
    struct _ap_copy_t <0>
{
    template <typename T>
        static void copy (T* dst, const T* src, unsigned cnt)
    {
        for (T* max = dst + cnt; dst < max; )
            *(dst++) = *(src++);
    }
};

}


/**
 *  @brief   A std::auto_ptr equivalent that works with heap arrays
 *  @author  Moritz Beutel
 *
 *
 *  This class is a very common implementation of std::auto_ptr,
 *  but it handles arrays instead of normal pointers.
 *
 *  The class provides support for dereferencing with the * and ->
 *  operators and indicing with [].
 */

template <typename T>
class ArrayPtr
{
private:
    class refc_t
    {
    private:
        T*          ptr;
        unsigned    refs;

        refc_t (T* p)
         : ptr (p), refs (1)
        {}
        ~refc_t (void)
        { delete [] ptr; }

    public:
        static refc_t* createRefC (T* p)
        { return p ? new refc_t (p) : 0; }
        void addRef (void)
        { ++refs; }
        void removeRef (void)
        {
            if ((--refs) == 0)
                delete this;
        }

        T* getPtr (void)
        { return ptr; }
        unsigned getRefCnt (void) const
        { return refs; }
    };

    refc_t* _ref;
    unsigned _size;

public:
    typedef T element_type;
    explicit ArrayPtr (element_type* ptr = NULL, unsigned size = ~0)
     : _ref (refc_t::createRefC (ptr)), _size (size) {}
    ArrayPtr (const ArrayPtr& rhs)
     : _ref (rhs._ref), _size (rhs._size)
    {
        if (_ref)
            _ref->addRef ();
    }

    ArrayPtr& operator = (const ArrayPtr& rhs)
    {
        if (_ref)
            _ref->removeRef ();
        _size = rhs._size;
        _ref = rhs._ref;
        if (_ref)
            _ref->addRef ();
        return *this;
    }

    ~ArrayPtr (void)
    {
        if (_ref)
            _ref->removeRef ();
    }

    element_type& operator [] (unsigned idx) const
    {
        _UCL_ASSERT (_ref, std::runtime_error, "ucl::stl::ArrayPtr: Invalid pointer");
        _UCL_ASSERT (idx < _size, std::out_of_range, "ucl::stl::ArrayPtr: Index out of range");
        return (_ref->getPtr ())[idx];
    }
    element_type& operator * (void) const
    {
        _UCL_ASSERT (_ref, std::runtime_error, "ucl::stl::ArrayPtr: Invalid pointer");
        return *_ref->getPtr ();
    }
    element_type* operator -> (void) const
    {
        _UCL_ASSERT (_ref, std::runtime_error, "ucl::stl::ArrayPtr: Invalid pointer");
        return _ref->getPtr ();
    }
    element_type* get (void) const
    {
        _UCL_ASSERT (_ref, std::runtime_error, "ucl::stl::ArrayPtr: Invalid pointer");
        return _ref->getPtr ();
    }
    ArrayPtr release (void)
    {
        ArrayPtr r = *this;
        if (_ref)
            _ref->removeRef ();
        _ref = 0;
        return r;
    }
    unsigned size (void) const
    { return _size; }
    void reset (element_type* ptr = NULL, unsigned size = ~0)
    {
        if (_ref)
            _ref->removeRef ();
        _ref = refc_t::createRefC (ptr);
        _size = size;
    }
    void alloc (unsigned size)
    {
        if (_ref)
            _ref->removeRef ();
        _ref = refc_t::createRefC (new T[size]);
        _size = size;
    }
    void copy (const ArrayPtr& rhs)
    {
        if (!rhs._ref)
            return;
        _UCL_ASSERT (rhs._size != ~0, std::runtime_error, "ucl::stl::ArrayPtr: Unknown size");
        alloc (rhs._size);
        detail::_ap_copy_t <IsPODType <T>::value>::copy (_ref->getPtr (), rhs._ref->getPtr (), _size);
    }

    unsigned ref_count (void) const
    { return _ref->getRefCnt (); }
};

} // namespace stl

} // namespace ucl


#endif // _UCL_ARRAYPTR_HPP
