
/*
 *  ucl/stl/arraywrap.hpp
 *
 *  Container class that wraps an array
 *  Author: Moritz Beutel
 */

#ifndef _UCL_ARRAYWRAP_HPP
#define _UCL_ARRAYWRAP_HPP

#include <cstdlib>
#include <string>
#include <algorithm>
#include <ucl/stl/meta.hpp>
#include <ucl/stl/iterwrap.hpp>


namespace ucl
{

namespace stl
{


/**
 *  @brief   A container class that wraps an array
 *  @author  Moritz Beutel
 *
 *
 *  This container class can be used as wrapper for a static or dynamic array.
 *
 *  The template parameters are:
 *   - <b>typename T</b> - the value type
 */

template <typename T>
    class ArrayWrapper
{
    typedef ArrayWrapper <T> Base;

protected:
    T* _array;
    std::size_t _size;

    void _throwOutOfRange (size_t offs)
    {
        char buf1[33], buf2[33];
        std::itoa (offs, buf1, 10);
        std::itoa (_size, buf2, 10);
        _UCL_DEBUG_THROW (std::out_of_range, std::string (
              "ucl::stl::ArrayWrapper: Index out of range. Index: ")
            + std::string (buf1) + ", size: " + buf2);
    }

public:
    typedef T value_type;
    typedef T* pointer;
    typedef const T* const_pointer;
    typedef T& reference;
    typedef const T& const_reference;
    typedef RandomIterator <T, Base, UCL_ITERWRAP_FORWARD>       iterator;
    typedef RandomIterator <const T, Base, UCL_ITERWRAP_FORWARD> const_iterator;
    typedef RandomIterator <T, Base, UCL_ITERWRAP_REVERSE>       reverse_iterator;
    typedef RandomIterator <const T, Base, UCL_ITERWRAP_REVERSE> const_reverse_iterator;
    typedef size_t size_type;
    typedef size_t difference_type;

        /// Returns an iterator that points to the first element of the array
    iterator begin (void)
    { return iterator (_array _UCL_ITERWRAP_CONSTRUCT); }
        /// Returns an const_iterator that points to the first element of the array
    const_iterator begin (void) const
    { return const_iterator (_array _UCL_ITERWRAP_CONSTRUCT); }
        /// Returns an const_iterator that points beyond the last element of the array
        /// Returns an iterator that points beyond the last element of the array
    iterator end (void)
    { return iterator (_array + _size _UCL_ITERWRAP_CONSTRUCT); }
    const_iterator end (void) const
    { return const_iterator (_array + _size _UCL_ITERWRAP_CONSTRUCT); }

        /// Returns an reverse_iterator that points to the last element of the array
    reverse_iterator rbegin (void)
    { return reverse_iterator (_array + _size - 1 _UCL_ITERWRAP_CONSTRUCT); }
        /// Returns an const_reverse_iterator that points to the last element of the array
    const_reverse_iterator rbegin (void) const
    { return const_reverse_iterator (_array + _size - 1 _UCL_ITERWRAP_CONSTRUCT); }
        /// Returns an reverse_iterator that points beyond the first element of the array
    reverse_iterator rend (void)
    { return reverse_iterator (_array - 1 _UCL_ITERWRAP_CONSTRUCT); }
        /// Returns an const_reverse_iterator that points beyond the first element of the array
    const_reverse_iterator rend (void) const
    { return const_reverse_iterator (_array - 1 _UCL_ITERWRAP_CONSTRUCT); }

        /// Returns a reference to the last element of the array
    reference back (void) { return _array[size - 1]; }
        /// Returns a const_reference to the last element of the array
    const_reference back (void) const { return _array[size - 1]; }
        /// Returns a reference to the first element of the array
    reference front (void) { return *_array; }
        /// Returns a const_reference to the first element of the array
    const_reference front (void) const { return *_array; }

        /// Swaps the array wrapper with another one
    void swap (Base& rhs)
    {
        std::swap (_array, rhs._array;
        std::swap (_size, rhs._size;
    }

        /// Returns the size of the array.
    size_type size (void)
    { return _size; }

        /// Empty arrays are not supported, therefore this function returns always false.
    bool empty (void) { return false; }

public:
        /** Constructs an array wrapper from a static array. The size is
         *  determined automatically.
         */
    template <size_type N> ArrayWrapper (T(&array)[N])
     : _size (N), _array (array) { StaticAssert <N != 0> sa; }

        /** Constructs an array wrapper from a dynamic array. The array size
         *  must be specified explicitly.
         */
    ArrayWrapper (T* array, size_type size)
     : _size (size), _array (array)
    {
        _UCL_DEBUG_STATEMENT (if (!size)) // some compilers allow this
            _UCL_THROW (std::range_error, "Zero length arrays are not supported!");
    }

    reference at (size_type offs)
    {
        if (offs >= _size)
            _throwOutOfRange (offs);
        return _array[offs];
    }
    const_reference at (size_type offs) const
    {
        if (offs >= _size)
            _throwOutOfRange (offs);
        return _array[offs];
    }
    reference operator [] (size_type offs)
    {
        _UCL_DEBUG_STATEMENT (if (offs >= _size))
            _UCL_DEBUG_STATEMENT (_throwOutOfRange (offs));
        return _array[offs];
    }
    const_reference operator [] (size_type offs) const
    {
        _UCL_DEBUG_STATEMENT (if (offs >= _size))
            _UCL_DEBUG_STATEMENT (_throwOutOfRange (offs));
        return _array[offs];
    }
};



} // namespace stl

} // namespace ucl

#endif // _UCL_ARRAYWRAP_HPP
