
/*
 *  ucl/stl/iterwrap.hpp
 *
 *  Iterator wrapper classes
 *  Author: Moritz Beutel
 *
 *  This file contains iterator wrappers that can be used for various purposes:
 *  - to wrap a plain pointer into an iterator
 *  - to wrap another iterator into an iterator
 *  - to wrap an iterator or pointer that does not provide the correct type
 */

#ifndef _UCL_ITERWRAP_HPP
#define _UCL_ITERWRAP_HPP

#include <iterator>
#include <cstring>
#include <cstddef>

#include <ucl/_debug.hpp>

#include <ucl/stl/meta.hpp>


namespace ucl
{

namespace stl
{


#define UCL_ITERWRAP_FORWARD           0
#define UCL_ITERWRAP_REVERSE           1
#define UCL_ITERWRAP_VIRTUAL_REVERSE   3


#ifdef UCL_ITERATOR_DEBUGGING
 #define _UCL_ITERWRAP_CONSTRUCT     , this
#else
 #define _UCL_ITERWRAP_CONSTRUCT
#endif


    /// @cond UCL_DETAIL
namespace detail
{

template <typename T, int direction>
    struct _iter_revwrap      // for iterators and pointers (forward)
{
    static void increment (T& i)
    { ++i; }
    static void decrement (T& i)
    { --i; }
    static bool smaller (T& lhs, T& rhs)
    { return (lhs < rhs); }
    static bool greater (T& lhs, T& rhs)
    { return (lhs > rhs); }
    enum { factor = 1 };
};
template <typename T>
    struct _iter_revwrap <T*, UCL_ITERWRAP_REVERSE> // for pointers
{
    static void increment (T*& i)
    { --i; }
    static void decrement (T*& i)
    { ++i; }
    static bool smaller (T*& lhs, T*& rhs)
    { return (lhs > rhs); }
    static bool greater (T*& lhs, T*& rhs)
    { return (lhs < rhs); }
    enum { factor = -1 };
};


template <int direction>
    struct _iter_helpers;
template <>
    struct _iter_helpers <UCL_ITERWRAP_FORWARD>
{
    template <typename Base>
        static typename Base::const_iterator getBegin (const Base* base)
    { return base->begin (); }
    template <typename Base>
        static typename Base::const_iterator getEnd (const Base* base)
    { return base->end (); }
    template <typename Base>
        static typename Base::iterator getBegin (Base* base)
    { return base->begin (); }
    template <typename Base>
        static typename Base::iterator getEnd (Base* base)
    { return base->end (); }

};
template <>
    struct _iter_helpers <UCL_ITERWRAP_REVERSE>
{
    template <typename Base>
        static typename Base::const_reverse_iterator getBegin (const Base* base)
    { return base->rbegin (); }
    template <typename Base>
        static typename Base::const_reverse_iterator getEnd (const Base* base)
    { return base->rend (); }
    template <typename Base>
        static typename Base::reverse_iterator getBegin (Base* base)
    { return base->rbegin (); }
    template <typename Base>
        static typename Base::reverse_iterator getEnd (Base* base)
    { return base->rend (); }
};

} // namespace detail
    /// @endcond



/**
 *  @brief   An bidirectional iterator wrapper class that wraps another
 *           iterator or pointer; iterator debugging is supported
 *  @author  Moritz Beutel
 *
 *
 *  This bidirectional iterator class can be used to wrap another operator
 *  or a plain pointer into an iterator.
 *  Iterator debugging (validity check, range check, base comparison) is
 *  activated if the macro UCL_ITERATOR_DEBUGGING is defined. Please refer to
 *  the documentation of the ucl::debug namespace for more information.
 *  The validity checks require an intact iterator to be passed to the constructor!
 *
 *  The template parameters are:
 *   - <b>typename T</b> - the value type
 *   - <b>typename Base</b> - the type of the container that owns the iterator
 *     type. Required for iterator debugging.
 *   - <b>int direction = UCL_ITERWRAP_FORWARD</b> - the iterator direction.
 *     Possible values are:
 *       -# UCL_ITERWRAP_FORWARD - forward iterator
 *       -# UCL_ITERWRAP_REVERSE - reverse iterator
 *   - <b>typename iter = T*</b> - the type of the wrapped iterator. Per default a
 *     plain pointer.
 *     The wrapped iterator does not need to result in T when dereferencing.
 *     The address of the dereferenced value is reinterpret_casted to T* if
 *     necessary.
 *   - <b>typename tag = std::bidirectional_iterator_tag</b> - the iterator tag.
 *     Use this only if you derive from BiDiIterator.
 */

template <typename T,
          typename Base,
          int direction = UCL_ITERWRAP_FORWARD,
          typename iter = T*,
          typename tag = std::bidirectional_iterator_tag> // only for derived classes
    class BiDiIterator
        : public std::iterator <tag, T>
{
    typedef std::iterator <tag, T> _iterBase;
private:
    struct _some_struct_t {};
public:
#ifdef UCL_BROKEN_META
    typedef _some_struct_t const_iterator_type; // no constructor for const_iterators
#else
    typedef typename IfThenElseType <TypesAreEqual <T, const T>::result,
        BiDiIterator <const T, Base, direction, iter>,
        _some_struct_t>::type
            const_iterator_type;
#endif
    typedef BiDiIterator iterator_type;
    typedef size_t size_type;

protected:
    iter _ptr;

#ifdef UCL_ITERATOR_DEBUGGING
    Base* _base;

    virtual std::string _getErrorPrefix (void) const
    { return (std::string (typeid (*this).name ()) + ": "); }
    virtual void _checkValidity (void) const
    {
        if (!(&*_ptr)
         || (detail::_iter_helpers <direction>::getEnd (_base) == *this))
            _UCL_ITERDBG_THROW (std::out_of_range, _getErrorPrefix () + "Invalid pointer");
    }
    void _checkInc (void)
    { _checkValidity (); }
    void _checkDec (void)
    {
        if (!(&*_ptr))
            _UCL_ITERDBG_THROW (std::out_of_range, _getErrorPrefix () + "Invalid pointer");
        if (detail::_iter_helpers <direction>::getBegin (_base) == *this)
            // leaves on the front side
            _UCL_ITERDBG_THROW (std::out_of_range, _getErrorPrefix () + "Invalid pointer");
    }
    void _checkBase (const iterator_type& rhs) const
    {
        if (_base != rhs._base)
            _UCL_ITERDBG_THROW (std::invalid_argument, _getErrorPrefix ()
                + "Iterators do not point to the same container");
    }
#endif


public:
    const iter& _getWrappedIterator (void) const { return _ptr; }

#ifdef UCL_ITERATOR_DEBUGGING
    BiDiIterator (void) : _ptr (NULL), _base (NULL) {}
    BiDiIterator (iter ptr, Base* base)
     : _ptr (ptr), _base (base)
    { } // no validity check - could point to base.[r]end ()
    BiDiIterator (const iterator_type& rhs)
     : _ptr (rhs._ptr), _base (rhs._base)
    { } // ditto
#else
    BiDiIterator (void) : _ptr (NULL) {}
    BiDiIterator (iter ptr) : _ptr (ptr) {}
    BiDiIterator (const iterator_type& rhs) : _ptr (rhs._ptr) {}
    BiDiIterator (const const_iterator_type& rhs) : _ptr (rhs._getWrappedIterator ()) {}
#endif

    _iterBase::pointer operator -> (void) const
    {
        _UCL_ITERDBG_STATEMENT (_checkValidity ());
        return reinterpret_cast <_iterBase::pointer> (&*_ptr);
    }
    _iterBase::reference operator * (void) const
    {
        _UCL_ITERDBG_STATEMENT (_checkValidity ());
        return *reinterpret_cast <_iterBase::pointer> (&*_ptr);
    }


    iterator_type& operator -- (void)
    {
        _UCL_ITERDBG_STATEMENT (_checkDec ());
        detail::_iter_revwrap <iter, direction>::decrement (_ptr);
        //--_ptr;
        return *this;
    }
    iterator_type operator -- (int)
    {
        _UCL_ITERDBG_STATEMENT (_checkDec ());
        iterator_type it = *this;
        detail::_iter_revwrap <iter, direction>::decrement (_ptr);
        //_ptr--;
        return it;
    }
    iterator_type& operator ++ (void)
    {
        _UCL_ITERDBG_STATEMENT (_checkInc ());
        detail::_iter_revwrap <iter, direction>::increment (_ptr);
        //++_ptr;
        return *this;
    }
    iterator_type operator ++ (int)
    {
        _UCL_ITERDBG_STATEMENT (_checkInc ());
        iterator_type it = *this;
        detail::_iter_revwrap <iter, direction>::increment (_ptr);
        //_ptr++;
        return it;
    }

    const iterator_type& operator = (const iterator_type& rhs)
    {
        _ptr = rhs._ptr;
        _UCL_ITERDBG_STATEMENT (_base = rhs._base);
    }

    bool operator == (const iterator_type& rhs) const
    {
        _UCL_ITERDBG_STATEMENT (_checkBase (rhs));
        return (_ptr == rhs._ptr);
    }
    bool operator != (const iterator_type& rhs) const
    {
        _UCL_ITERDBG_STATEMENT (_checkBase (rhs));
        return (_ptr != rhs._ptr);
    }
};




/**
 *  @brief   An random iterator wrapper class that wraps another
 *           iterator or pointer; iterator debugging is supported
 *  @author  Moritz Beutel
 *
 *
 *  This random access iterator class can be used to wrap another operator
 *  or a plain pointer into an iterator.
 *  Iterator debugging (validity check, range check, base comparison) is
 *  activated if the macro UCL_ITERATOR_DEBUGGING is defined. Please refer to
 *  the documentation of the ucl::debug namespace for more information.
 *
 *  The template parameters are:
 *   - <b>typename T</b> - the value type
 *   - <b>typename Base</b> - the type of the container that owns the iterator
 *     type. Required for iterator debugging.
 *   - <b>int direction = UCL_ITERWRAP_FORWARD</b> - the iterator direction.
 *     Possible values are:
 *       -# UCL_ITERWRAP_FORWARD - forward iterator
 *       -# UCL_ITERWRAP_REVERSE - reverse iterator
 *   - <b>typename iter = T*</b> - the type of the wrapped iterator. Per default a
 *     plain pointer.
 *     The wrapped iterator does not need to result in T when dereferencing.
 *     The address of the dereferenced value is reinterpret_casted to T* if
 *     necessary.
 */

template <typename T,
          typename Base,
          int direction = UCL_ITERWRAP_FORWARD,
          typename iter = T*>
    class RandomIterator
        : public BiDiIterator <T, Base, direction, iter, std::random_access_iterator_tag>
{
    typedef std::iterator <std::random_access_iterator_tag, T> _iterBase;
    typedef BiDiIterator <T, Base, direction, iter> _iterBase2;
private:
    struct _some_struct_t {};

public:
#ifdef UCL_BROKEN_META
    typedef _some_struct_t const_iterator_type; // no constructor for const_iterators
#else
    typedef typename IfThenElseType <TypesAreEqual <T, const T>::result,
        RandomIterator <const T, Base, direction, iter>,
        _some_struct_t>::type
            const_iterator_type;
#endif
    typedef RandomIterator iterator_type;
    typedef size_t size_type;


#ifdef UCL_ITERATOR_DEBUGGING
protected:
    void _checkOffset (int offs)
    {
        if ((detail::_iter_revwrap <iter, direction>::smaller (_ptr + offs,
             detail::_iter_helpers <direction>::getBegin (_base)))
         || (detail::_iter_revwrap <iter, direction>::greater (_ptr + offs,
             detail::_iter_helpers <direction>::getEnd (_base))))
            _UCL_ITERDBG_THROW (std::out_of_range, _getErrorPrefix () + "Pointer out of range");
    }
    void _checkRange (void) const
    { _checkOffset (0); }
    virtual void _checkValidity (void) const
    {
        _checkRange ();
        _iterBase2::_checkValidity ();
    }
#endif

public:
#ifdef UCL_ITERATOR_DEBUGGING
    RandomIterator (void) : _iterBase2 (NULL, NULL) {}
    RandomIterator (iter ptr, Base* base)
     : _iterBase2 (ptr, base)
    { _UCL_ITERDBG_STATEMENT (_checkRange ()); }
    RandomIterator (const iterator_type& rhs)
     : _iterBase2 (rhs._ptr, rhs._base)
    { _UCL_ITERDBG_STATEMENT (_checkRange ()); }
#else
    RandomIterator (void) : _iterBase2 (NULL) {}
    RandomIterator (iter ptr) : _iterBase2 (ptr) {}
    RandomIterator (const iterator_type& rhs) : _iterBase2 (rhs._ptr) {}
    RandomIterator (const const_iterator_type& rhs) : _iterBase2 (rhs._getWrappedIterator ()) {}
#endif

    _iterBase::reference operator [] (size_type offs) const
    {
        offs *= detail::_iter_revwrap <iter, direction>::factor;
        _checkRange (offs);
        return _ptr[offs];
    }

    const iterator_type& operator = (const iterator_type& rhs)
    {
        _ptr = rhs._ptr;
        _UCL_ITERDBG_STATEMENT (_base = rhs._base);
    }
    bool operator == (const iterator_type& rhs) const
    {
        _UCL_ITERDBG_STATEMENT (_checkBase (rhs));
        return (_ptr == rhs._ptr);
    }
    bool operator != (const iterator_type& rhs) const
    {
        _UCL_ITERDBG_STATEMENT (_checkBase (rhs));
        return (_ptr != rhs._ptr);
    }
    bool operator <= (const iterator_type& rhs) const
    {
        _UCL_ITERDBG_STATEMENT (_checkBase (rhs));
        return detail::_iter_revwrap <iter, direction>::greater (rhs._ptr, _ptr);
    }
    bool operator >= (const iterator_type& rhs) const
    {
        _UCL_ITERDBG_STATEMENT (_checkBase (rhs));
        return detail::_iter_revwrap <iter, direction>::smaller (rhs._ptr, _ptr);
    }
    bool operator < (const iterator_type& rhs) const
    {
        _UCL_ITERDBG_STATEMENT (_checkBase (rhs));
        return detail::_iter_revwrap <iter, direction>::smaller (_ptr, rhs._ptr);
    }
    bool operator > (const iterator_type& rhs) const
    {
        _UCL_ITERDBG_STATEMENT (_checkBase (rhs));
        return detail::_iter_revwrap <iter, direction>::greater (_ptr, rhs._ptr);
    }


    _iterBase::difference_type operator - (const iterator_type& rhs) const
    {
        _UCL_ITERDBG_STATEMENT (_checkBase (rhs));
        return _ptr - rhs._ptr;
    }
    iterator_type operator - (_iterBase::difference_type offs) const
    {
        offs *= detail::_iter_revwrap <iter, direction>::factor;
        _UCL_ITERDBG_STATEMENT (_checkRange (-offs));
        return iterator_type (_ptr - offs, _base);
    }
    iterator_type operator + (_iterBase::difference_type offs) const
    {
        offs *= detail::_iter_revwrap <iter, direction>::factor;
        _UCL_ITERDBG_STATEMENT (_checkRange (offs));
        return iterator_type (_ptr + offs, _base);
    }
    iterator_type& operator -= (_iterBase::difference_type offs)
    {
        offs *= detail::_iter_revwrap <iter, direction>::factor;
        _UCL_ITERDBG_STATEMENT (_checkRange (-offs));
        _ptr -= offs;
        return (*this);
    }
    iterator_type& operator += (_iterBase::difference_type offs)
    {
        offs *= detail::_iter_revwrap <iter, direction>::factor;
        _UCL_ITERDBG_STATEMENT (_checkRange (offs));
        _ptr += offs;
        return (*this);
    }
};



} // namespace stl

} // namespace ucl

#endif // _UCL_ITERWRAP_HPP
