
/*
 *  ucl/stl/meta.hpp
 *
 *  Meta Template Definitions
 *  Author: Moritz Beutel
 *
 *  This file contains various definitions of meta templates
 *  that are used in other parts of this library.
 *  There may be some intersection with other popular libraries
 *  like Boost, but in some cases it is better if a library does
 *  not depend on another library, especially if it required just
 *  such a minor part.
 */

#ifndef _UCL_META_HPP
#define _UCL_META_HPP


#ifdef __BCPLUSPLUS__
 #if __BCPLUSPLUS__ < 0x0560 // C++Builder 5 and before
  #define UCL_BROKEN_META
 #endif
#endif

namespace ucl
{

namespace stl
{

/**
 *  @brief   Static assertion template
 *  @author  Moritz Beutel
 *
 *  This template raises an assertion if the given value is false.
 *  Use it like this:
 *
 *  <c>static const ucl::StaticAssert <some_condition>;</c>
 */
#ifdef _DOXYGEN
template <bool cond> class StaticAssert {};
#else
template <bool cond> class StaticAssert;
#endif

    /// @cond UCL_DETAIL
template <> class StaticAssert <true> {};
    /// @endcond



/**
 *  @brief   Inverses the current argument
 *  @author  Moritz Beutel
 *
 *  This template inverts its argument. The inverted argument
 *  is stored in the member result:
 *
 *  <c>const not_1 = ucl::Not <1>::result;</c>
 */
template <bool cond> struct Not
{ enum { result = cond ? 0 : 1 }; };


/**
 *  @brief   Determines whether two types are equal
 *  @author  Moritz Beutel
 *
 *  This template determines whether the two given types
 *  are equal. The member result is 1 if they are equal, 0 otherwise.
 *
 *  <c>bool isMSVC6 = TypesAreEqual <wchar_t, unsigned short>::result;</c>
 */
template <typename T1, typename T2> struct TypesAreEqual
{ enum { result = 0 }; };

    /// @cond UCL_DETAIL
template <typename T> struct TypesAreEqual <T, T>
{ enum { result = 1 }; };
    /// @endcond


/**
 *  @brief   A type selection depending on a static condition
 *  @author  Moritz Beutel
 *
 *  <c>typedef typename ucl::IfThenElseType <TypesAreEqual <wchar_t, unsigned short>::result,
 *                                        myOwnWCharTEmulatingClass,
 *                                        wchar_t> myWCharT;</c>
 */
template <bool cond, typename T1, typename T2> struct IfThenElseType
{ typedef T1 type; };

    /// @cond UCL_DETAIL
template <typename T1, typename T2> struct IfThenElseType <false, T1, T2>
{ typedef T2 type; };
    /// @endcond

} // namespace stl

} // namespace ucl


#endif // _UCL_META_HPP
