
/*
 *  ucl/stl/smartptr.hpp
 *
 *  A common smart pointer implementation
 *  Author: Moritz Beutel
 */

#ifndef _UCL_SMARTPTR_HPP
#define _UCL_SMARTPTR_HPP

#include <ucl/_debug.hpp>
#include <stdexcept>


namespace ucl
{

namespace stl
{


/**
 *  @brief   A common smart pointer implementation
 *  @author  Moritz Beutel
 *
 *
 *  A common smart pointer implementation
 */

template <typename T>
class SmartPtr
{
private:
    class refc_t
    {
    private:
        T*          ptr;
        unsigned    refs;

        refc_t (T* p)
         : ptr (p), refs (1)
        {}
        ~refc_t (void)
        { delete ptr; }

    public:
        static refc_t* createRefC (T* p)
        { return p ? new refc_t (p) : 0; }
        void addRef (void)
        { ++refs; }
        void removeRef (void)
        {
            if ((--refs) == 0)
                delete this;
        }

        T* getPtr (void)
        { return ptr; }
        unsigned getRefCnt (void) const
        { return refs; }
    };

    refc_t* _ref;

    
public:
    typedef T element_type;

    explicit SmartPtr (element_type* ptr = NULL)
     : _ref (refc_t::createRefC (ptr)) {}
    SmartPtr (const SmartPtr& rhs)
     : _ref (rhs._ref)
    {
        if (_ref)
            _ref->addRef ();
    }

    SmartPtr& operator = (const SmartPtr& rhs)
    {
        if (_ref)
            _ref->removeRef ();
        _ref = rhs._ref;
        if (_ref)
            _ref->addRef ();
        return *this;
    }

    ~SmartPtr (void)
    {
        if (_ref)
            _ref->removeRef ();
    }

    element_type& operator * (void) const
    {
        _UCL_ASSERT (_ref, std::runtime_error, "ucl::stl::ArrayPtr: Invalid pointer");
        return *_ref->getPtr ();
    }
    element_type* operator -> (void) const
    {
        _UCL_ASSERT (_ref, std::runtime_error, "ucl::stl::ArrayPtr: Invalid pointer");
        return _ref->getPtr ();
    }
    element_type* get (void) const
    {
        return _ref ? _ref->getPtr () : 0;
    }
    SmartPtr release (void)
    {
        SmartPtr r = *this;
        if (_ref)
            _ref->removeRef ();
        _ref = 0;
        return r;
    }
    void reset (element_type* ptr = NULL)
    {
        if (_ref)
            _ref->removeRef ();
        _ref = refc_t::createRefC (ptr);
    }
    void alloc (void)
    {
        if (_ref)
            _ref->removeRef ();
        _ref = refc_t::createRefC (new T);
    }
    void copy (const SmartPtr& rhs)
    {
        if (_ref)
            _ref->removeRef ();
        if (rhs._ref)
            _ref = refc_t::createRefC (new T (*(rhs._ref->getPtr ())));
        else
            _ref = 0;
    }

    unsigned ref_count (void) const
    { return _ref->getRefCnt (); }
};

} // namespace stl

} // namespace ucl


#endif // _UCL_SMARTPTR_HPP
