
#ifndef _UCL_THREAD_HPP
#define _UCL_THREAD_HPP


#include <windows.h>
#include <ucl/stl/arrayptr.hpp>
#include <ucl/stl/smartptr.hpp>
#include <string>
#include <list>
#include <vector>

namespace ucl
{

namespace win
{


class BasicThread
{
private:
    HANDLE hThread;
    DWORD dwThreadId;
    bool running, cancelled;

    static DWORD WINAPI ThreadFunc (LPVOID data);
    
protected:
    virtual DWORD execute (void) = 0;

public:
    BasicThread (SECURITY_ATTRIBUTES* sa = 0);
    virtual ~BasicThread (void);

    bool isActive (void) const
    { return (hThread != NULL); }
    bool isRunning (void) const
    { return running; }
    bool isCancelled (void) const
    { return cancelled; }

    void suspend (void);
    void resume (void);
    void cancel (void);
    void terminate (void);
    void reset (void);

    void setPriority (int priority)
    { SetThreadPriority (hThread, priority); }
    int getPriority (void) const
    { return GetThreadPriority (hThread); }

    HANDLE getHandle (void)
    { return hThread; }
    DWORD getThreadId (void)
    { return dwThreadId; }

    DWORD getExitCode (void) const;
};



class ThreadPool
{
    class WorkThread;
    friend class WorkThread;

public:
    class Task;
    friend class Task;

    static const unsigned msgBusy           = 0x00000001;
    static const unsigned msgIdle           = 0x00000002;
    static const unsigned msgTaskBegin      = 0x00000003;
    static const unsigned msgTaskCompleted  = 0x00000004;
    static const unsigned msgTaskTerminated = 0x00000005;
    static const unsigned msgException      = 0x00000006;

    typedef void (*taskfunc_t) (void* obj, Task& task);
    typedef void (*msgfunc_t)  (void* data, unsigned msg, ucl::stl::SmartPtr <Task> task);

private:
    class WorkThread : public BasicThread
    {
        friend class ThreadPool;

    private:
        ucl::stl::SmartPtr <Task> task;
        ThreadPool* owner;
        unsigned    index;
        HANDLE      hEvent;
        bool        waiting;

    protected:
        virtual DWORD execute (void);

    public:
        WorkThread (ThreadPool* o, unsigned idx);
        virtual ~WorkThread (void);
    };
    
    typedef ucl::stl::SmartPtr <WorkThread> PWorkThread;
    typedef ucl::stl::SmartPtr <Task> PTask;

    struct msgdata_t
    {
        HWND hWnd;
        UINT uMsg;

        msgdata_t (HWND wnd, UINT msg)
         : hWnd (wnd), uMsg (msg)
        {}
    };


        /* 0x0000001 - free msgdata */
    unsigned            flags;
    unsigned            threads_active;

    msgfunc_t           msgfunc;
    void*               msgdata;

    CRITICAL_SECTION    cs;
    //HANDLE              hEvent;

    std::vector <PWorkThread> threads;
    std::list   <PTask>       tasks;



    static void _MsgHandler (void* data, unsigned msg, ucl::stl::SmartPtr <Task> task);
    template <class C>
    struct _taskhandler_t
    {
        template <void (C::*Method) (Task*)>
            static void _TaskHandler (void* data, Task& task)
        { (reinterpret_cast <C*> (data)->*Method) (task); }
    };

    void _init (unsigned thread_cnt);
    void _terminateTask (Task* task);
    void _cancelPending (void);

public:
    class Task
    {
        friend class ThreadPool;
        friend class WorkThread;

    public:
        enum taskstatus_t
        {
            tsPending = 0,
            tsBusy,
            tsCompleted,
            tsFailed,
            tsTerminated,
        };

    private:
        taskfunc_t      func;
        void*           obj;
        void*           data;
        unsigned        threadidx;
        taskstatus_t    status;
        ThreadPool*     owner;
        std::string     msg;
        std::list <PTask>::iterator iter; 

        Task (ThreadPool* o, taskfunc_t _func, void* _obj, void* _data)
         : func (_func), obj (_obj), data (_data), status (tsPending)
        {}

    public:
        taskstatus_t getStatus (void) const     { return status; }
        void*        getTaskData (void)         { return data; }
        const void*  getTaskData (void) const   { return data; }
        void*        getTaskObj (void)          { return obj; }
        const void*  getTaskObj (void) const    { return obj; }
        const std::string& getErrorMessage (void) const { return msg; }

            // if you kill a single task, you have to cleanup the data manually!
        void terminate (void) { owner->_terminateTask (this); }
    };


    ThreadPool (msgfunc_t msg_func, void* msg_data, unsigned thread_cnt = 0);
    ThreadPool (HWND hWnd, UINT uMsg, unsigned thread_cnt = 0);
    ~ThreadPool (void);

    ucl::stl::SmartPtr <Task> addTask (taskfunc_t taskfunc, void* obj, void* taskdata = 0);
    template <class C, void (C::*Method) (Task&)>
        ucl::stl::SmartPtr <Task> addTask (C& obj, void* taskdata = 0)
    { return addTask (&_taskhandler_t <C>::_TaskHandler <Method>, &obj, taskdata); }

    unsigned activeThreads (void) const
    { return threads_active; }

    void terminateAllTasks (void);
    void cancelPendingTasks (void);
};

} // namespace win

} // namespace ucl

#endif // _UCL_THREAD_HPP
